#include "flacProvider.hpp"

#include <string.h>

#include <algorithm>
#include <functional>
#include <iostream>
#include <mutex>
#include <thread>

#include "../timing.hpp"
#include "../utils.hpp"
#include "flacDecoder.hpp"

#define PLAYBACK_SAMPLERATE 48000

#define PRELOADED_DATA_MILLISECONDS 5000
#define PRELOADED_DATA_MINIMUM_MILLISECONDS 3000

void FlacProvider::preprocessData()
{
    while (flacDecoder.rawDataLength < flacDecoder.rawDataAllocatedLength && flacDecoder.unreadSamplesLeft != 0)
    {
        if (flacDecoder.getMoreData() == false)
        {
            // TODO: handle error in getting more data
            return;
        }
    }

    while (true)
    {
        long inputFrames = std::min(flacDecoder.rawDataAllocatedLength - flacDecoder.rawDataReadPosition, flacDecoder.rawDataLength) / 2;
        long outputFrames = std::min(preloadedDataAllocatedLength - preloadedDataWritePosition, preloadedDataAllocatedLength - preloadedDataLength) / 2;

        resamplerData.end_of_input = (inputFrames == flacDecoder.rawDataLength && flacDecoder.unreadSamplesLeft == 0);
        resamplerData.data_in = flacDecoder.rawData + flacDecoder.rawDataReadPosition;
        resamplerData.data_out = preloadedDataBuffer + preloadedDataWritePosition;
        resamplerData.input_frames = inputFrames;
        resamplerData.output_frames = outputFrames;
        resamplerData.src_ratio = resamplingRatio;

        int processError = src_process(resamplerState, &resamplerData);
        if (processError != 0)
        {
            std::cerr << "error resampling data: " << src_strerror(processError) << std::endl;
        }

        flacDecoder.rawDataLength -= resamplerData.input_frames_used * 2;
        preloadedDataLength += resamplerData.output_frames_gen * 2;
        flacDecoder.rawDataReadPosition += resamplerData.input_frames_used * 2;
        if (flacDecoder.rawDataReadPosition == flacDecoder.rawDataAllocatedLength)
        {
            flacDecoder.rawDataReadPosition = 0;
        }
        preloadedDataWritePosition += resamplerData.output_frames_gen * 2;
        if (preloadedDataWritePosition == preloadedDataAllocatedLength)
        {
            preloadedDataWritePosition = 0;
        }

        // TODO: problems may accour if data is actually left but there was no space left to process more data
        if (resamplerData.input_frames_used == 0)
        {
            if (flacDecoder.unreadSamplesLeft == 0)
            {
                unconvertedAudioLeft = false;
            }
            break;
        }
    }
}

// it is assumed that sampleCount is divisible by 2 (2 channel audio)
void FlacProvider::getData(int16_t *audioData, size_t sampleCount, size_t sampleIndex, int64_t samplePosition)
{
    // TODO: add total samples played from provider so music playback can start from the middle of the track
    // not so important currently

    int64_t playbackStartTime = serverTimeToLocalTime(startTime);
    int64_t sampleCountInMicroseconds = sampleCount * 1000000 / 2 / PLAYBACK_SAMPLERATE;

    // code to offset audio playback start to sync with playback start time that has been set previously
    size_t i = 0;
    if (samplePosition + sampleCountInMicroseconds <= playbackStartTime)
    {
        return;
    }
    else if (samplePosition < playbackStartTime)
    {
        i += (samplePosition + sampleCountInMicroseconds - playbackStartTime) * PLAYBACK_SAMPLERATE * 2 / 1000000;
    }

    for (; i < sampleCount; i += 2)
    {
        if (preloadedDataLength == 0)
        {
            if (!unconvertedAudioLeft)
            {
                // reached to the end of the audio file
                shouldRemove = true;
            }
            else
            {
                std::cerr << "ran out of preloaded samples before music ended" << std::endl;
            }
            break;
        }
        audioData[i] = sat_adds16b(audioData[i], preloadedDataBuffer[preloadedDataReadPosition] * 32767);
        audioData[i + 1] = sat_adds16b(audioData[i + 1], preloadedDataBuffer[preloadedDataReadPosition + 1] * 32767);

        preloadedDataLength -= 2;
        preloadedDataReadPosition += 2;
        if (preloadedDataReadPosition == preloadedDataAllocatedLength)
        {
            preloadedDataReadPosition = 0;
        }
    }
}

bool FlacProvider::loadFile(std::string filepath)
{
    FLAC__StreamDecoderInitStatus init_status = flacDecoder.init(filepath);
    if (init_status != FLAC__STREAM_DECODER_INIT_STATUS_OK)
    {
        std::cerr << "error initializing flac decoder: " << FLAC__StreamDecoderInitStatusString[init_status] << std::endl;
        return false;
    }

    if (flacDecoder.process_until_end_of_metadata() == false)
    {
        FLAC__StreamDecoderState state = flacDecoder.get_state();
        std::cerr << "error processing metadata: " << FLAC__StreamDecoderStateString[state] << std::endl;
        return false;
    }
    resamplingRatio = (double)PLAYBACK_SAMPLERATE / flacDecoder.sampleRate;

    std::cout << "samplerate: " << flacDecoder.sampleRate << std::endl;

    return initResampler();
}

bool FlacProvider::initResampler()
{
    int err;
    resamplerState = src_new(RESAMPLER_TYPE, 2, &err);
    if (resamplerState == nullptr)
    {
        std::cerr << "error initializing resampler: " << src_strerror(err) << std::endl;
        return false;
    }
    return true;
}

FlacProvider::FlacProvider(std::string filepath, std::chrono::milliseconds _startTime, std::function<void()> _onEndedCallback)
    : startTime(_startTime), flacDecoder(), onEndedCallback(_onEndedCallback)
{
    preloadedDataAllocatedLength = 2 * PRELOADED_DATA_MILLISECONDS * PLAYBACK_SAMPLERATE / 1000;
    preloadedDataBuffer = new float[preloadedDataAllocatedLength];

    if (loadFile(filepath) == false)
    {
        delete[] preloadedDataBuffer;
        throw std::runtime_error("failed loading file: " + filepath);
    }

    preprocessingThread = std::thread([this]() {
        while (true)
        {
            preprocessData();

            {
                std::lock_guard<std::mutex> lock(preprocessingMutex);
                if (!preprocessingFinished)
                {
                    preprocessingFinished = true;
                    preprocessingConditionVariable.notify_all();
                }
            }

            if (shouldExitPreprocessingThread)
            {
                break;
            }
            if (shouldRemove)
            {
                std::function<void()> onEndedCallbackClone = onEndedCallback;
                std::thread([onEndedCallbackClone]() {
                    onEndedCallbackClone();
                }).detach();
                break;
            }

            std::this_thread::sleep_for(std::chrono::milliseconds(100));
        }
    });
}

FlacProvider::~FlacProvider()
{
    shouldExitPreprocessingThread = true;
    preprocessingThread.join();
    delete[] preloadedDataBuffer;
}

void FlacProvider::waitForInitialPreprosessing()
{
    std::unique_lock<std::mutex> lock(preprocessingMutex);
    preprocessingConditionVariable.wait(lock, [this]() {
        return preprocessingFinished;
    });
}