#include "audioSystem.hpp"

#include <math.h>
#include <pulse/pulseaudio.h>
#include <stdio.h>

#include <algorithm>
#include <chrono>
#include <condition_variable>
#include <iostream>
#include <mutex>
#include <thread>
#include <vector>

#include "audioProviders/flacProvider.hpp"
#include "timing.hpp"

// start

// TODO: check chrono clock resolution

// TODO: cleanup sample bytesize mess

static std::thread audioPlaybackThread;

static std::mutex mainThreadMutex;
static std::condition_variable mainThreadConditionVariable;
static bool mainThreadShouldContinue = false;

static std::mutex audioProviderPlaybackThreadMutex;
// stop

// TODO: make global
#define PLAYBACK_SAMPLERATE 48000
#define PLAYBACK_SAMPLE_BYTESIZE sizeof(int16_t)

int playbackBufferSize = 80000;

static std::vector<AudioProvider *> audioProviders;

static pa_mainloop *pa_ml;
static pa_mainloop_api *pa_mlapi;
static pa_context *pa_ctx;
static pa_stream *playstream;
static pa_sample_spec playbackSpec;
static pa_buffer_attr bufferAttributes;

static int pa_ready = 0;

void addAudioProvider(AudioProvider *provider) // this is actually completely deadlock safe
{
    audioProviderPlaybackThreadMutex.lock();
    audioProviders.push_back(provider);
    audioProviderPlaybackThreadMutex.unlock();
}

void removeAudioProvider(AudioProvider *provider)
{
    audioProviderPlaybackThreadMutex.lock();
    audioProviders.erase(std::remove(audioProviders.begin(), audioProviders.end(), provider), audioProviders.end());
    audioProviderPlaybackThreadMutex.unlock();
}

void shutdownAudioSystem()
{
    /*
    shouldShutdown = true;
    if (playbackInProgress)
    {
        stopPlayback();
    }
    pa_mainloop_quit(pa_ml, 0);
    */
    audioPlaybackThread.join();
}

#define PLAYBACK_TIME_POSITION_ROUND_COUNT 50
int playbackTimePositionIndex = 0;

bool calibratePlaybackTiming = false;
bool playbackTimeCalibrated = false;

// TODO: make sure these are atomic

// in microseconds
int64_t playbackStartTime = 0;
// in microseconds
int64_t playbackStartTimeSteadyClock = 0;
// in microseconds
int64_t playbackStartTimeCalibrationMinMaxDifference = 0; // TODO: fix variable name

std::vector<int64_t> playbackStartPoints(PLAYBACK_TIME_POSITION_ROUND_COUNT);

static void timingInfoCB(pa_stream *s, int success, void *userdata)
{
    pa_usec_t usec;
    int neg;

    const pa_timing_info *timingInfo = pa_stream_get_timing_info(playstream);

    if (timingInfo == nullptr)
        return;
    /*
    std::cout << "configured_sink_usec   " << timingInfo->configured_sink_usec << "\n";
    std::cout << "configured_source_usec " << timingInfo->configured_source_usec << "\n";
    std::cout << "playing                " << timingInfo->playing << "\n";
    std::cout << "read_index             " << timingInfo->read_index << "\n";
    std::cout << "read_index_corrupt     " << timingInfo->read_index_corrupt << "\n";
    std::cout << "since_underrun         " << timingInfo->since_underrun << "\n";
    std::cout << "sink_usec              " << timingInfo->sink_usec << "\n";
    std::cout << "source_usec            " << timingInfo->source_usec << "\n";
    std::cout << "synchronized_clocks    " << timingInfo->synchronized_clocks << "\n";
    std::cout << "timestamp.tv_sec       " << timingInfo->timestamp.tv_sec << "\n";
    std::cout << "timestamp.tv_usec      " << timingInfo->timestamp.tv_usec << "\n";
    std::cout << "transport_usec         " << timingInfo->transport_usec << "\n";
    std::cout << "write_index            " << timingInfo->write_index << "\n";
    std::cout << "write_index_corrupt    " << timingInfo->write_index_corrupt << std::endl;
    */

    if (timingInfo->playing)
    {
        playbackStartPoints[playbackTimePositionIndex] = (int64_t)timingInfo->timestamp.tv_sec * 1000000 + timingInfo->timestamp.tv_usec - pa_bytes_to_usec(timingInfo->read_index, &playbackSpec);

        playbackTimePositionIndex++;
        if (playbackTimePositionIndex == PLAYBACK_TIME_POSITION_ROUND_COUNT)
        {
            /*
            std::cout << "normal clock start time: " << playbackStartTime << std::endl;
            std::cout << "steady clock start time: " << playbackStartTimeSteadyClock << std::endl
                      << std::endl;
                      */
            playbackTimePositionIndex = 0;
            playbackTimeCalibrated = true;
        }
        if (playbackTimeCalibrated)
        {
            int64_t playbackStartTimeRounding = 0;
            int64_t minTime = LONG_LONG_MAX;
            int64_t maxTime = LONG_LONG_MIN;
            for (auto playbackStartPoint : playbackStartPoints) // TODO: maybe add checking for if local time has been adjusted by checking if values are too different
            {
                minTime = std::min(minTime, playbackStartPoint);
                maxTime = std::max(maxTime, playbackStartPoint);
                playbackStartTimeRounding += playbackStartPoint / PLAYBACK_TIME_POSITION_ROUND_COUNT;
            }
            playbackStartTime = playbackStartTimeRounding;
            playbackStartTimeCalibrationMinMaxDifference = maxTime - minTime;

            // FIXME: add system to check if local time has jumped

            auto currentRealTime = std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::system_clock::now().time_since_epoch());
            auto currentSteadyTime = std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::steady_clock::now().time_since_epoch());

            int64_t microsecondsSincePlaybackStarted = currentRealTime.count() - playbackStartTime;
            playbackStartTimeSteadyClock = currentSteadyTime.count() - microsecondsSincePlaybackStarted;
        }
    }
    /*
    int getLatencyReturn = pa_stream_get_latency(playstream, &usec, &neg);

    if (getLatencyReturn == -PA_ERR_NODATA)
    {
        printf("nodata\n");
    }
    else
    {
        printf("latency %lu us\n", usec);
        std::cout << "data left: " << timingInfo->write_index - timingInfo->read_index << std::endl;
    }
    */
}

static size_t getAudio(int16_t *audioData, size_t sampleCount, size_t sampleIndex, int64_t samplePosition)
{
    std::fill(audioData, audioData + sampleCount, 0);

    audioProviderPlaybackThreadMutex.lock();
    for (AudioProvider *audioProvider : audioProviders)
    {
        audioProvider->getData(audioData, sampleCount, sampleIndex, samplePosition); // TODO: better handling of saturated addition
    }
    audioProviderPlaybackThreadMutex.unlock();

    return sampleCount;
}

static size_t audioWriteIndex = 0;

bool firstCalibrated = true;

static bool pulseSystemInitialized = false;

static void stream_request_cb(pa_stream *s, long unsigned int length, void *userdata)
{
    // std::cout << "asked for " << length / 2 << " samples of audio" << std::endl;
    // std::cout << "playbackBuffer size: " << pa_usec_to_bytes(playbackBufferSize, &playbackSpec) / 2 << " samples" << std::endl;

    pa_stream_update_timing_info(s, timingInfoCB, nullptr);

    if (!playbackTimeCalibrated || firstCalibrated)
    {
        int16_t *audioData;
        long unsigned int audioDataLength = length;

        pa_stream_begin_write(s, (void **)&audioData, &audioDataLength);
        // TODO: handle errors: https://freedesktop.org/software/pulseaudio/doxygen/stream_8h.html#a6cf50cfc4ea8897391941184d74d7dfa
        std::fill(audioData, audioData + audioDataLength / PLAYBACK_SAMPLE_BYTESIZE, 0);
        pa_stream_write(s, audioData, audioDataLength, nullptr, 0, PA_SEEK_RELATIVE);
        // TODO: handle errors: https://freedesktop.org/software/pulseaudio/doxygen/stream_8h.html#a4fc69dec0cc202fcc174125dc88dada7
        audioWriteIndex += audioDataLength;

        if (firstCalibrated && playbackTimeCalibrated)
        {
            firstCalibrated = false;
            calibratePlaybackTiming = false;
            mainThreadShouldContinue = true;
            mainThreadConditionVariable.notify_one();
        }

        return;
    }

    int16_t *audioData;
    long unsigned int audioDataLength = length;
    pa_stream_begin_write(s, (void **)&audioData, &audioDataLength);

    long unsigned int writtenDataLength = getAudio(audioData, audioDataLength / PLAYBACK_SAMPLE_BYTESIZE, audioWriteIndex / PLAYBACK_SAMPLE_BYTESIZE, playbackStartTimeSteadyClock + pa_bytes_to_usec(audioWriteIndex, &playbackSpec)) * PLAYBACK_SAMPLE_BYTESIZE;
    pa_stream_write(s, audioData, writtenDataLength, nullptr, 0, PA_SEEK_RELATIVE);
    audioWriteIndex += writtenDataLength;
}

static void stream_underflow_cb(pa_stream *s, void *userdata)
{
    std::cout << "stream underflow" << std::endl;
}

static void pa_state_cb(pa_context *c, void *userdata)
{
    pa_context_state_t state;
    state = pa_context_get_state(c);
    switch (state)
    {
        case PA_CONTEXT_UNCONNECTED:
            std::cout << "pa_state_cb: PA_CONTEXT_UNCONNECTED" << std::endl;
            break;
        case PA_CONTEXT_CONNECTING:
            std::cout << "pa_state_cb: PA_CONTEXT_CONNECTING" << std::endl;
            break;
        case PA_CONTEXT_AUTHORIZING:
            std::cout << "pa_state_cb: PA_CONTEXT_AUTHORIZING" << std::endl;
            break;
        case PA_CONTEXT_SETTING_NAME:
            std::cout << "pa_state_cb: PA_CONTEXT_SETTING_NAME" << std::endl;
            break;
        default:
            std::cout << "pa_state_cb: default" << std::endl;
            break;
        case PA_CONTEXT_FAILED:
            std::cout << "pa_state_cb: PA_CONTEXT_FAILED" << std::endl;
            pa_ready = 2;
            break;
        case PA_CONTEXT_TERMINATED:
            pa_ready = 2;
            std::cout << "pa_state_cb: PA_CONTEXT_TERMINATED" << std::endl;
            break;
        case PA_CONTEXT_READY:
            pa_ready = 1;
            std::cout << "pa_state_cb: PA_CONTEXT_READY" << std::endl;
            break;
    }
}

static bool initPulseSystem()
{
    playbackSpec.format = PA_SAMPLE_S16LE;
    playbackSpec.channels = 2;
    playbackSpec.rate = PLAYBACK_SAMPLERATE;

    pa_ml = pa_mainloop_new();
    pa_mlapi = pa_mainloop_get_api(pa_ml);
    pa_ctx = pa_context_new(pa_mlapi, "poppi2");

    pa_context_set_state_callback(pa_ctx, pa_state_cb, NULL);

    pa_context_connect(pa_ctx, NULL, PA_CONTEXT_NOFLAGS, NULL);
    while (pa_ready == 0)
    {
        pa_mainloop_iterate(pa_ml, 1, NULL);
    }

    if (pa_ready == 2)
    {
        std::cerr << "failed connecting to pulse server" << std::endl;
        pa_context_unref(pa_ctx);
        pa_mainloop_free(pa_ml);
        return false;
    }

    playstream = pa_stream_new(pa_ctx, "Playback", &playbackSpec, NULL);
    if (playstream == NULL)
    {
        std::cerr << "failed creating a new stream" << std::endl;
        pa_context_unref(pa_ctx);
        pa_mainloop_free(pa_ml);
        return false;
    }

    pa_stream_set_write_callback(playstream, stream_request_cb, NULL);
    pa_stream_set_underflow_callback(playstream, stream_underflow_cb, NULL);

    bufferAttributes.maxlength = pa_usec_to_bytes(playbackBufferSize, &playbackSpec);
    bufferAttributes.prebuf = pa_usec_to_bytes(playbackBufferSize, &playbackSpec);
    bufferAttributes.tlength = pa_usec_to_bytes(playbackBufferSize, &playbackSpec); // because PA_STREAM_ADJUST_LATENCY is set, the playback buffer size is automatically set to the same as this
    bufferAttributes.minreq = 0;

    if (pa_stream_connect_playback(playstream, NULL, &bufferAttributes, (pa_stream_flags_t)(PA_STREAM_ADJUST_LATENCY), NULL, NULL) < 0)
    {
        std::cerr << "failed connecting for playback" << std::endl;
        pa_stream_unref(playstream);
        pa_context_unref(pa_ctx);
        pa_mainloop_free(pa_ml);
        return false;
    }

    const pa_buffer_attr *actualBufferAttributes = nullptr;
    while (actualBufferAttributes == nullptr)
    {
        pa_mainloop_iterate(pa_ml, 1, NULL);
        actualBufferAttributes = pa_stream_get_buffer_attr(playstream);
    }

    bufferAttributes = *actualBufferAttributes;
    playbackBufferSize = pa_bytes_to_usec(bufferAttributes.maxlength, &playbackSpec);

    pulseSystemInitialized = true;
    return true;
}

static void audioPlaybackThreadFunction()
{
    std::cout << "start mainloop" << std::endl;
    while (true)
    {
        pa_mainloop_iterate(pa_ml, 1, NULL);
    }
}

static void startCalibratePlaybackTiming()
{
    calibratePlaybackTiming = true;
    std::unique_lock<std::mutex> lk(mainThreadMutex);
    mainThreadConditionVariable.wait(lk, [] { return mainThreadShouldContinue; });
    mainThreadShouldContinue = false;
}

bool startAudioSystem()
{
    if (initPulseSystem() == false)
    {
        std::cerr << "failed initializing pulse system" << std::endl;
        return false;
    }
    std::cout << "Initialized pulse system" << std::endl;

    audioPlaybackThread = std::thread(audioPlaybackThreadFunction);
    // audioPreprocessorThread = std::thread(audioPreprocessorThreadFunction);

    std::cout << "Created playback and preprocessing threads" << std::endl;

    startCalibratePlaybackTiming();
    std::cout << "calibration done" << std::endl;
    std::cout << "starting time: " << playbackStartTime << " with max difference of: " << playbackStartTimeCalibrationMinMaxDifference << " microseconds" << std::endl;

    return true;
}
