#include "crashHandler.hpp"

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <chrono>
#include <cstring>
#include <functional>
#include <iostream>
#include <thread>

bool isRestartableExit(int exitInfo)
{
    bool isRestartable = true;
    std::string basicExplanation;

    if (WIFEXITED(exitInfo))
    {
        int exitStatus = WEXITSTATUS(exitInfo);

        switch (exitStatus)
        {
            case EXIT_SUCCESS:
                isRestartable = false;
                break;
            default:
                basicExplanation = "Poppi exited with status code " + exitStatus;
                break;
        }
    }
    else if (WIFSIGNALED(exitInfo))
    {
        int terminationSignal = WTERMSIG(exitInfo);
        std::string signalName = strsignal(terminationSignal);

        switch (terminationSignal)
        {
            case SIGHUP:
            case SIGINT:
            case SIGQUIT:
            case SIGKILL: // I'm conflicted about if it should automatically restart on SIGKILL
            case SIGTERM:
                isRestartable = false;
            case SIGSEGV:
                basicExplanation = "Crash: Segmentation fault";
                break;
            case SIGABRT:
                basicExplanation = "Abort: Unhandled exception";
                break;
            default:
                basicExplanation = "Exited by signal \"" + signalName + "\" (" + std::to_string(terminationSignal) + ")";
                break;
        }
    }
    else
    {
        // Should not happen???? Application should always exit with a status or a signal
        // https://www.gnu.org/software/libc/manual/html_node/Process-Completion-Status.html
        std::cerr << "Poppi exited with mystery???? Was not WIFEXITED nor WIFSIGNALED!" << std::endl;
        abort();
    }

    if (isRestartable)
    {
        std::cerr << "\n\n\n"
                  << basicExplanation << std::endl;
    }

    return isRestartable;
}

#define BURST_RESTART_LIMIT 5
#define BURST_RESTART_TIME std::chrono::seconds(10)
#define AFTER_BURST_FAIL_RESTART_TIME std::chrono::minutes(1)

static int restartCount = 0;
static std::chrono::time_point<std::chrono::steady_clock> burstFirstRestartTime = std::chrono::steady_clock::now();

void runWithCrashHandler(ExecutionParameters params, std::function<int(ExecutionParameters)> mainFunction, std::function<void()> beforeRestart)
{
    bool firstRun = true;
    bool shouldRestart = true;

    while (shouldRestart)
    {
        pid_t pid = fork();
        if (pid == 0)
        {
            // Don't print the version number on the first run as it is already printed in the parent process
            if (!firstRun)
                beforeRestart();
            firstRun = false;

            int exitCode = mainFunction(params);
            exit(exitCode);
        }
        else if (pid < 0)
        {
            std::cerr << "Failed to start the child process: " << std::strerror(errno) << std::endl;
            abort();
        }

        int exitInfo;
        waitpid(pid, &exitInfo, 0);

        if (!isRestartableExit(exitInfo))
        {
            shouldRestart = false;
            continue;
        }

        std::cerr << "Process exited unexpectedly. ";

        std::chrono::time_point<std::chrono::steady_clock> currentTime = std::chrono::steady_clock::now();
        std::chrono::duration<double> timeSinceFirstRestart = currentTime - burstFirstRestartTime;

        if (timeSinceFirstRestart > BURST_RESTART_TIME)
        {
            restartCount = 0;
            burstFirstRestartTime = currentTime;
        }
        else
        {
            if (restartCount >= BURST_RESTART_LIMIT)
            {
                std::cerr << "It has crashed " << restartCount << " times in " << std::chrono::duration_cast<std::chrono::seconds>(BURST_RESTART_TIME).count() << " seconds. Waiting "
                          << std::chrono::duration_cast<std::chrono::seconds>(AFTER_BURST_FAIL_RESTART_TIME).count() << " seconds before restarting again." << std::endl;
                std::this_thread::sleep_for(AFTER_BURST_FAIL_RESTART_TIME);

                burstFirstRestartTime = currentTime;
            }
        }

        std::cerr << "Restarting...\n\n\n"
                  << std::endl;
        restartCount++;
    }
}
