#include "TCPClient.hpp"

#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <unistd.h>

#include <cstring>
#include <vector>

#include "../glibcPolyfills/strerrorname_np.hpp"

namespace {
std::string explainConnectError(int error, addrinfo* addressInfo)
{
    std::string errorName = strerrorname_np(error);
    std::string errorString = strerror(error);
    std::string errnoErrorString = "\"" + errorName + "\" " + errorString;

    std::string errorExplanation = "";
    switch (error)
    {
        case ECONNREFUSED:
            errorExplanation = "Host reachable, but poppi server not running or a firewall is blocking the connection.";
            break;
        case EHOSTUNREACH:
        case EHOSTDOWN:
            errorExplanation = "Usually means the host is down, doesn't exist or a firewall is blocking the connection. (Some router that received the original packet from its source/gateway port doesn't know how to route it forward because it doesn't exist on the network)";
            break;
        case ENETUNREACH:
            if (addressInfo->ai_family == AF_INET)
                errorExplanation = "Usually means that you are not connected to the internet. (The address is in another network and this network (usually you) doesn't have a default gateway to where route packets that are not in the same network)";
            else if (addressInfo->ai_family == AF_INET6)
                errorExplanation = "Since IPv6 is not widely used, this is likely because you don't have an IPv6 address. (The address is in another network and this network (usually you) doesn't have a default gateway to where route packets that are not in the same network)";
            break;
        case ETIMEDOUT:
            errorExplanation = "Usually means that the host down or a firewall is dropping all incoming connections without a response.";
            break;
    }
    std::string fullErrorString = "connect error: " + errnoErrorString;
    if (!errorExplanation.empty())
    {
        fullErrorString += "\n    " + errorExplanation;
    }

    return fullErrorString;
}

const char* get_gai_error_name(int error_code)
{
    switch (error_code)
    {
        case EAI_ADDRFAMILY:
            return "EAI_ADDRFAMILY";
        case EAI_AGAIN:
            return "EAI_AGAIN";
        case EAI_BADFLAGS:
            return "EAI_BADFLAGS";
        case EAI_FAIL:
            return "EAI_FAIL";
        case EAI_FAMILY:
            return "EAI_FAMILY";
        case EAI_MEMORY:
            return "EAI_MEMORY";
        case EAI_NODATA:
            return "EAI_NODATA";
        case EAI_NONAME:
            return "EAI_NONAME";
        case EAI_SERVICE:
            return "EAI_SERVICE";
        case EAI_SOCKTYPE:
            return "EAI_SOCKTYPE";
        case EAI_SYSTEM:
            return "EAI_SYSTEM";
        default:
            return "Unknown error";
    }
}

std::string explainResolveError(int errnoError, int status)
{
    std::string errorName;
    std::string errorString;

    std::string errorExplanation = "";

    if (status == EAI_SYSTEM)
    {
        errorName = strerrorname_np(errnoError);
        errorString = strerror(errnoError);

        // TODO: Figure out what errors can happen in normal ish situations and explain them.
        // switch (errnoError)
        // {
        // }
    }
    else
    {
        errorName = get_gai_error_name(status);
        errorString = gai_strerror(status);

        switch (status)
        {
            case EAI_AGAIN:
                errorExplanation = "Could not connect to name server. (Are you connected to the internet?)";
                break;
            case EAI_NONAME:
                errorExplanation = "Name does not exist. (Are you sure you spelled the address correctly?)";
                break;
        }
    }

    std::string errnoErrorString = "\"" + errorName + "\" " + errorString;

    std::string fullErrorString = "Name resolve error: " + errnoErrorString;
    if (!errorExplanation.empty())
    {
        fullErrorString += "\n    " + errorExplanation;
    }

    return fullErrorString;
}

struct addrinfo* resolveAddress(std::string host, std::string port)
{
    struct addrinfo hints = {};
    struct addrinfo* result;

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    int status = getaddrinfo(host.c_str(), port.c_str(), &hints, &result);
    if (status != 0)
    {
        int errnoError = errno;

        std::string connectErrorString = "Failed to resolve address \"" + host + "\"\n" + explainResolveError(errnoError, status);

        throw TCPClient::ConnectException(connectErrorString);
    }

    return result;
}

} // namespace

void TCPClient::connect(std::string hostAddress, std::string port)
{
    std::lock_guard lock(socketOperationMutex);

    if (socketState != SocketState::DISCONNECTED)
    {
        throw ConnectException("Socket not disconnected");
    }
    if (socketfd != -1)
    {
        throw std::logic_error("Socket file descriptor is not -1 but socket state is disconnected");
    }

    addrinfo* addressInfo = resolveAddress(hostAddress, port);

    // To my knowledge, this should never happen as resolveAddress should always return either a valid addrinfo struct or throw an exception.
    if (addressInfo == nullptr)
    {
        throw ConnectException("Unknown error when resolving address (addrinfo is nullptr)");
    }

    std::vector<std::pair<int, addrinfo*>> connectErrors;

    while (addressInfo != nullptr)
    {
        socketfd = socket(addressInfo->ai_family, addressInfo->ai_socktype, addressInfo->ai_protocol);
        if (socketfd < 0)
        {
            std::cerr << "Error when creating socket??: \"" << strerrorname_np(errno) << "\" " << strerror(errno) << std::endl;
            addressInfo = addressInfo->ai_next;
            continue;
        }

        if (::connect(socketfd, addressInfo->ai_addr, addressInfo->ai_addrlen) < 0)
        {
            int error = errno;
            connectErrors.push_back({error, addressInfo});
            addressInfo = addressInfo->ai_next;

            close(socketfd);
            socketfd = -1;

            continue;
        }

        break;
    }

    if (addressInfo == nullptr)
    {
        std::string connectErrorString = "Failed to connect to \"" + hostAddress + ":" + port + "\"";

        char resolvedIp[NI_MAXHOST];
        char resolvedPort[NI_MAXSERV];
        if (connectErrors.size() == 1)
        {
            addrinfo* errorAddressInfo = connectErrors[0].second;

            int getNameInfoError = getnameinfo(errorAddressInfo->ai_addr, errorAddressInfo->ai_addrlen, resolvedIp, NI_MAXHOST, resolvedPort, NI_MAXSERV, NI_NUMERICHOST | NI_NUMERICSERV);

            if (getNameInfoError != 0)
            {
                std::cerr << "Failed to get info about name info: " << gai_strerror(getNameInfoError) << std::endl;
                connectErrorString = ". Could not get address from internal address structure but was able to resolve it???\n";
            }
            else
            {
                connectErrorString += " resolved as \"" + std::string(resolvedIp) + "\"\n";
            }
            connectErrorString += explainConnectError(connectErrors[0].first, errorAddressInfo);
        }
        else
        {
            connectErrorString += ". Errors for all resolved addresses:";
            for (auto& [connectErrno, errorAddressInfo] : connectErrors)
            {
                connectErrorString += "\n";
                int getNameInfoError = getnameinfo(errorAddressInfo->ai_addr, errorAddressInfo->ai_addrlen, resolvedIp, NI_MAXHOST, resolvedPort, NI_MAXSERV, NI_NUMERICHOST | NI_NUMERICSERV);

                if (getNameInfoError != 0)
                {
                    std::cerr << "Failed to get info about name info: " << gai_strerror(getNameInfoError) << std::endl;
                    connectErrorString += "    Could not get address from internal address structure but was able to resolve it???\n    ";
                }
                else
                {
                    connectErrorString += "    \"" + std::string(resolvedIp) + "\": ";
                }
                connectErrorString += explainConnectError(connectErrno, errorAddressInfo);
            }
        }

        throw ConnectException(connectErrorString);
    }

    freeaddrinfo(addressInfo);

    int setsockoptFlag = 1;
    int setsocketoptError = setsockopt(socketfd, IPPROTO_TCP, TCP_NODELAY, &setsockoptFlag, sizeof(int));
    if (setsocketoptError < 0)
    {
        std::cerr << "ALERT! Failed to set TCP_NODELAY: \"" << strerrorname_np(errno) << "\" " << strerror(errno) << "\n";
        std::cerr << "This might cause sync issues." << std::endl;
    }

    socketState = SocketState::CONNECTED;
}

TCPClient::~TCPClient()
{
    if (socketfd != -1)
    {
        close(socketfd);
    }
}

// Sends a FIN to the server and waits for the server to send a FIN back.
// If the server doesn't send a FIN back, the server is forcefully disconnected.
// Never exits until the socket is disconnected.
void TCPClient::disconnect()
{
    using namespace std::chrono_literals;

    // Since this waits for the data receiving thread to exit, the socket operation mutex is unlocked
    // with the state set to DISCONNECTING. The only allowed functions to run in this state is the receive function.
    std::unique_lock lock(socketOperationMutex);

    if (socketState == SocketState::DISCONNECTED)
    {
        return;
    }

    if (socketState == SocketState::DISCONNECTING)
    {
        if (finishedCV.wait_for(lock, 60s, [&] { return socketState == SocketState::DISCONNECTED; }))
        {
            return;
        }

        throw std::runtime_error("Data receiving thread did not exit in 60 seconds after disconnecting the socket via another thread.");
    }

    socketState = SocketState::DISCONNECTING;

    // Try gracefully disconnecting
    if (shutdown(socketfd, SHUT_WR) < 0)
    {
        int errnoError = errno;
        if (errnoError != ENOTCONN)
        {
            throw std::runtime_error("Failed to shutdown socket: " + std::string(strerrorname_np(errnoError)) + ": " + strerror(errnoError));
        }
    }

    if (finishedCV.wait_for(lock, 10s, [&] { return socketState == SocketState::DISCONNECTED; }))
    {
        return;
    }

    std::cerr << "Failed to receive FIN from server in 10 seconds. Forcing disconnect." << std::endl;

    // Graceful disconnect failed, forcefully disconnect
    disconnectedForcefully = true;
    if (shutdown(socketfd, SHUT_RDWR) < 0)
    {
        int errnoError = errno;
        if (errnoError != ENOTCONN)
        {
            throw std::runtime_error("Failed to shutdown socket: " + std::string(strerrorname_np(errnoError)) + ": " + strerror(errnoError));
        }
    }

    if (finishedCV.wait_for(lock, 20s, [&] { return socketState == SocketState::DISCONNECTED; }))
    {
        return;
    }

    throw std::runtime_error("Data receiving thread did not exit in 20 seconds after forcefully disconnecting the socket.");
}

void TCPClient::send(const char* data, size_t size, bool corc)
{
    std::lock_guard lock(socketOperationMutex);

    if (socketState != SocketState::CONNECTED)
    {
        throw SendException("Socket not connected");
    }

    while (size > 0)
    {
        int sendFlags = MSG_NOSIGNAL;
        if (corc)
        {
            sendFlags |= MSG_MORE;
        }
        ssize_t sent = ::send(socketfd, data, size, sendFlags);
        if (sent < 0)
        {
            // TODO: Better description of errors
            int errnoError = errno;

            std::string errorString = strerrorname_np(errnoError) + std::string(": ") + strerror(errnoError);

            throw SendException("Failed to send data: " + errorString);
        }

        data += sent;
        size -= sent;
    }
}

// A failed receive will result in the socket being forcefully disconnected.
size_t TCPClient::receive(char* data, size_t size)
{
    if (!socketReceiveMutex.try_lock())
    {
        throw std::logic_error("Only one thread can receive data at a time");
    }
    std::lock_guard lock(socketReceiveMutex, std::adopt_lock);

    {
        std::lock_guard<std::mutex> operationLock(socketOperationMutex);

        if (!(socketState == SocketState::CONNECTED || socketState == SocketState::DISCONNECTING))
        {
            throw NotConnectedException("Socket not connected");
        }
    }

    ssize_t received = ::recv(socketfd, data, size, 0);
    if (received <= 0)
    {
        int errnoError = errno;
        std::string errorString = strerrorname_np(errnoError) + std::string(": ") + strerror(errnoError);

        std::lock_guard<std::mutex> operationLock(socketOperationMutex);

        close(socketfd);
        socketfd = -1;
        socketState = SocketState::DISCONNECTED;
        finishedCV.notify_all();

        if (received < 0)
        {
            throw ReceiveException("Failed to receive data: " + errorString);
        }
    }

    return received;
}
