#include "messageHandler.hpp"

#include <arpa/inet.h>

#include <chrono>
#include <cstdio>
#include <cstring>
#include <fstream>
#include <iostream>
#include <stdexcept>
#include <string>
#include <thread>

#include "../fileSystem.hpp"
#include "../playbackSystem/playbackSystem.hpp"
#include "networking.hpp"
#include "protocolEnums.hpp"

using namespace networking;

void onMessageReceived(RecieveMessageType messageType, unsigned char *message, uint16_t length)
{
    switch (messageType)
    {
        case RecieveMessageType::SET_PLAYBACK_START_TIME:
        {
            // 8 bytes playback start time in mills
            // +1 because filename has to be at least 1 characters
            if (length < 8 + 1)
            {
                std::cerr << "recieved too short message" << std::endl;
                break;
            }
            uint64_t id = ntohll(*reinterpret_cast<uint64_t *>(message));
            auto playbackStartTime = std::chrono::milliseconds(ntohll(*reinterpret_cast<uint64_t *>(message + 8)));
            std::string filename = std::string(reinterpret_cast<char *>(message + 16), length - 16);

            // Might be a problem that addNewMusic and removeMusic do block for a few moments.
            // What makes this hard with multithreading, is that addNewMusic and removeMusic needs to be called in the same order as they are recieved.
            try
            {
                addNewMusic(id, filename, playbackStartTime);
            }
            catch (std::runtime_error &e)
            {
                std::cerr << "failed to add new music: " << e.what() << std::endl;
            }

            break;
        }
        case RecieveMessageType::STOP_PLAYBACK_TIME:
        {
            // 8 bytes id
            if (length < 8)
            {
                std::cerr << "recieved too short message" << std::endl;
                break;
            }
            uint64_t id = ntohll(*reinterpret_cast<uint64_t *>(message));
            removeMusic(id);
            break;
        }
        case RecieveMessageType::INIT_FILE:
        {
            // packet format:
            // n bytes filename (not null terminated)
            if (length < 1)
            {
                std::cerr << "recieved too short message" << std::endl;
                break;
            }
            std::string filename(reinterpret_cast<char *>(message), length);
            std::cout << "removing file: " << filename << std::endl;
            std::remove(("music/" + filename).c_str());
            break;
        }
        case RecieveMessageType::FILE_CHUNK:
        {
            // packet format:
            // 1 byte filename length
            // n bytes filename (not null terminated)
            // m bytes file data
            // if packet length is same as filename length + 1, then this is the last packet

            if (length < 2)
            {
                std::cerr << "recieved too short message" << std::endl;
                break;
            }
            uint8_t filenameLength = message[0];
            if (length < 1 + filenameLength)
            {
                std::cerr << "recieved too short message" << std::endl;
                break;
            }
            std::string filename(reinterpret_cast<char *>(message + 1), filenameLength);

            if (length == filenameLength + 1)
            {
                std::cout << "got last chunk for file: " << filename << std::endl;
                try
                {
                    networking::sendMessage(SendMessageType::FILE_DONE, reinterpret_cast<unsigned char *>(const_cast<char *>(filename.c_str())), filename.length());
                }
                catch (networking::NotConnectedException &e)
                {
                    std::cerr << "failed to send FILE_DONE: " << e.what() << std::endl;
                }
            }
            else
            {
                appendToFile(filename, message + 1 + filenameLength, length - 1 - filenameLength);
            }

            break;
        }
        case RecieveMessageType::GET_ALL_FILES:
        {
            // TODO: figure out how to handle stuff like this when a client is suddenly disconnected
            std::thread([]() {
                std::vector<std::string> files = getFiles();

                uint32_t fileCount = files.size();
                uint32_t fileCountNetwork = htonl(fileCount);
                try
                {
                    networking::sendMessage(SendMessageType::FILE_LIST_INIT, reinterpret_cast<unsigned char *>(&fileCountNetwork), sizeof(fileCount));
                }
                catch (NotConnectedException &e)
                {
                    std::cerr << "failed to send FILE_LIST_INIT: " << e.what() << std::endl;
                    return;
                }

                for (auto &file : files)
                {
                    std::string hash = getFileHash(file);

                    unsigned char *data = new unsigned char[file.length() + hash.length() + 1];

                    memcpy(data, file.c_str(), file.length());
                    data[file.length()] = 0;
                    memcpy(data + file.length() + 1, hash.c_str(), hash.length());

                    try
                    {
                        networking::sendMessage(SendMessageType::FILE_LIST_ENTRY, data, file.length() + hash.length() + 1);
                    }
                    catch (NotConnectedException &e)
                    {
                        std::cerr << "failed to send FILE_LIST_ENTRY: " << e.what() << std::endl;
                        delete[] data;
                        return;
                    }
                    delete[] data;
                }

                try
                {
                    networking::sendMessage(SendMessageType::FILE_LIST_DONE);
                }
                catch (NotConnectedException &e)
                {
                    std::cerr << "failed to send FILE_LIST_DONE: " << e.what() << std::endl;
                }
            }).detach();
            break;
        }
        case RecieveMessageType::MOVE_FILE:
        {
            // packet format:
            // 1 byte filename length
            // n bytes filename (not null terminated)
            // 1 byte new filename length
            // m bytes new filename (not null terminated)
            if (length < 4)
            {
                std::cerr << "recieved too short message" << std::endl;
                break;
            }
            uint8_t filenameLength = message[0];
            if (length < 1 + filenameLength)
            {
                std::cerr << "recieved too short message" << std::endl;
                break;
            }
            std::string filename(reinterpret_cast<char *>(message + 1), filenameLength);

            uint8_t newFilenameLength = message[1 + filenameLength];
            if (length < 1 + filenameLength + 1 + newFilenameLength)
            {
                std::cerr << "recieved too short message" << std::endl;
                break;
            }
            std::string newFilename(reinterpret_cast<char *>(message + 1 + filenameLength + 1), newFilenameLength);

            std::cout << "moving file: " << filename << " to " << newFilename << std::endl;
            moveFile("music/" + filename, "music/" + newFilename);

            try
            {
                networking::sendMessage(SendMessageType::MOVE_FILE);
            }
            catch (NotConnectedException &e)
            {
                std::cerr << "failed to send MOVE_FILE: " << e.what() << std::endl;
            }
            break;
        }
        default:
            break;
    }
}