#include "networking.hpp"

#include <arpa/inet.h>

#include <atomic>
#include <condition_variable>
#include <cstring>
#include <fstream>
#include <iostream>
#include <mutex>
#include <sstream>
#include <thread>

#include "../timing.hpp"
#include "TCPClient.hpp"
#include "messageHandler.hpp"
#include "protocolEnums.hpp"

#define HEADER_SIZE 3
#define PROTOCOL_PREAMBLE "poppi"
#define CACHED_DIFFERENCE_COUNT 20
#define PONG_COUNT_BEFORE_CONSIDERED_CONNECTED 40
#define PING_INTERVAL std::chrono::milliseconds(300)

namespace {
std::mutex sendPacketMutex;

enum ProtocolMessageType
{
    PING = 0,
    PONG = 1,
    MESSAGE = 2,
    CLOCKS_SYNCED = 3,
};

std::string serverAdderss, serverPort;

std::thread connectionThread;
std::thread clockSyncThread;

TCPClient *tcpClient;

std::mutex clocksSyncedMutex;
std::condition_variable clocksSyncedCV;

std::mutex protocolInitializedMutex;
bool protocolInitialized = false;

std::chrono::milliseconds latestRecievedServerTime;
std::atomic<std::chrono::time_point<std::chrono::steady_clock>> clockSyncMessageHighResRecieveTime;

bool clocksSynced = false;

void clockSyncThreadFunction();
void connectionThreadFunction();
void sendProtocolMessage(ProtocolMessageType type, unsigned char *message, uint16_t length);
bool disconnectFromServer();

void sendProtocolMessage(ProtocolMessageType type, unsigned char *message, uint16_t length)
{
    {
        std::lock_guard<std::mutex> lk(protocolInitializedMutex);
        if (!protocolInitialized)
            throw TCPClient::NotConnectedException("not connected to server");
    }
    std::lock_guard<std::mutex> lock(sendPacketMutex);

    uint8_t header[HEADER_SIZE];
    header[0] = type;
    uint16_t *header_messageLength = (uint16_t *)&header[1];
    *header_messageLength = htons(length + HEADER_SIZE);

    tcpClient->send((char *)&header, HEADER_SIZE);
    if (length != 0)
    {
        tcpClient->send((char *)message, length);
    }
}

void sendProtocolMessage(ProtocolMessageType type)
{
    sendProtocolMessage(type, nullptr, 0);
}

#define TEMP_MESSAGE_TIMEOUT std::chrono::milliseconds(10000)

std::atomic<std::chrono::time_point<std::chrono::steady_clock>> latestRecievedMessageTime;
std::condition_variable syncThreadNotificationVariable;
std::mutex syncThreadShouldContinueMutex;
bool syncThreadShouldContinue = false;

class PongTimeoutException : public std::runtime_error
{
  public:
    PongTimeoutException(const std::string &message)
        : std::runtime_error(message) {}
};

class ClientNotConnectedException : public std::runtime_error
{
  public:
    ClientNotConnectedException(const std::string &message)
        : std::runtime_error(message) {}
};

void waitForPongMessage()
{
    // this system waits for up to TEMP_MESSAGE_TIMEOUT for a pong message to be recieved but also resets
    // the timer if a message is recieved. This is to prevent the system from false timeouts if for some
    // reason the pong message is stuck behind slowly processed other messages
    latestRecievedMessageTime = std::chrono::steady_clock::now();

    bool cvWaitForExitedByNotify = false;

    std::unique_lock<std::mutex> lk(syncThreadShouldContinueMutex);
    while (true)
    {
        auto syncWaitTime = latestRecievedMessageTime.load() - std::chrono::steady_clock::now() + TEMP_MESSAGE_TIMEOUT;
        if (syncWaitTime < std::chrono::milliseconds(0))
            break;

        cvWaitForExitedByNotify = syncThreadNotificationVariable.wait_for(lk, syncWaitTime, [&] {
            return syncThreadShouldContinue;
        });
        if (cvWaitForExitedByNotify)
            break;
    }
    syncThreadShouldContinue = false;

    std::lock_guard<std::mutex> protocolInitLock(protocolInitializedMutex);
    if (!cvWaitForExitedByNotify)
    {
        throw PongTimeoutException("pong message timed out");
    }
    else if (!protocolInitialized)
    {
        throw ClientNotConnectedException("Not connected to server or connected, but connection not initialized");
    }
}

void clockSyncThreadFunction()
{
    struct ClockDifference
    {
        std::chrono::microseconds difference;
        std::chrono::microseconds oneWayLatency;
        std::chrono::time_point<std::chrono::steady_clock> timeRecieved;
    };

    ClockDifference differences[CACHED_DIFFERENCE_COUNT];

    for (int i = 0; i < CACHED_DIFFERENCE_COUNT; i++)
    {
        // TODO: Fix this. This is stupid spaghetty initialization
        differences[i].difference = std::chrono::milliseconds(0);
        differences[i].oneWayLatency = std::chrono::seconds(1000);
        differences[i].timeRecieved = std::chrono::steady_clock::now() - std::chrono::seconds(1000);
    }

    uint64_t pongRecieveCount = 0;

    while (true)
    {
        {
            std::unique_lock<std::mutex> lk(protocolInitializedMutex);
            if (!protocolInitialized)
            {
                std::cout << "waiting for protocol to be initialized" << std::endl;
                clocksSynced = false;
                pongRecieveCount = 0;

                syncThreadNotificationVariable.wait(lk, [&] { return protocolInitialized; });
                std::cout << "protocol initialized" << std::endl;
            }
        }

        {
            // reset wait status
            std::lock_guard<std::mutex> lk(syncThreadShouldContinueMutex);
            syncThreadShouldContinue = false;
        }
        auto clockSyncMessageHighResSendTime = std::chrono::steady_clock::now();
        try
        {
            sendProtocolMessage(ProtocolMessageType::PING);
        }
        catch (TCPClient::NotConnectedException &e)
        {
            continue;
        }

        try
        {
            waitForPongMessage();
        }
        catch (PongTimeoutException &e)
        {
            std::cout << "failed to recieve pong (timeout)" << std::endl;
            disconnectFromServer();
            continue;
        }
        catch (ClientNotConnectedException &e)
        {
            continue;
        }
        pongRecieveCount++;

        std::chrono::microseconds oneWayLatency = std::chrono::duration_cast<std::chrono::microseconds>(clockSyncMessageHighResRecieveTime.load() - clockSyncMessageHighResSendTime) / 2;

        // TODO: detect if diff suddenly shiftes a lot and alert the user
        std::chrono::microseconds clockDiff = std::chrono::duration_cast<std::chrono::microseconds>(clockSyncMessageHighResRecieveTime.load().time_since_epoch()) - latestRecievedServerTime + oneWayLatency;

        // Rank differences by how recent they are and how accurate the latency is.
        // Drift per hour is just arbitrarily assumed to be around what it can be in the worst case. It should rather be slightly too high than too low.
        // Too low drift causes it to retain old data for too long causing the application clock to drift. Too high drift causes it to discard data too quickly resulting in jittery clock
        // and not as accurate clock as it prefers newer ping results that had higher latency.
        std::chrono::microseconds clockDriftPerHour = std::chrono::milliseconds(10);

        std::size_t leastAccurateIndex = 0;
        std::chrono::microseconds leastAccurateDifference = std::chrono::milliseconds(0);
        for (size_t i = 0; auto &difference : differences)
        {
            std::chrono::microseconds timeSinceRecieve = std::chrono::duration_cast<std::chrono::microseconds>(clockSyncMessageHighResRecieveTime.load() - difference.timeRecieved);
            std::chrono::microseconds clockDrift = std::chrono::microseconds(static_cast<int64_t>(timeSinceRecieve.count() * (static_cast<float>(clockDriftPerHour.count()) / 60 / 60 / 1000 / 1000)));

            if (difference.oneWayLatency + clockDrift > leastAccurateDifference)
            {
                leastAccurateDifference = difference.oneWayLatency + clockDrift;
                leastAccurateIndex = i;
            }
            i++;
        }

        // If worst drift compensated latency is still better than the new one, then the new one is discarded
        // Only if we are on exactly the PONG_COUNT_BEFORE_CONSIDERED_CONNECTED pong count, we allow the new one to be used
        // to not make the system wait for the next better difference before reporting that it is connected
        if (oneWayLatency > leastAccurateDifference && pongRecieveCount != PONG_COUNT_BEFORE_CONSIDERED_CONNECTED)
        {
            if (pongRecieveCount >= PONG_COUNT_BEFORE_CONSIDERED_CONNECTED)
                std::this_thread::sleep_for(PING_INTERVAL);
            continue;
        }

        // Update the least accurate difference with the new one
        differences[leastAccurateIndex].difference = clockDiff;
        differences[leastAccurateIndex].oneWayLatency = oneWayLatency;
        differences[leastAccurateIndex].timeRecieved = clockSyncMessageHighResRecieveTime.load();

        // If we have received a total of at least CACHED_DIFFERENCE_COUNT pongs, then we can consider the connection to be stable
        if (pongRecieveCount >= PONG_COUNT_BEFORE_CONSIDERED_CONNECTED)
        {
            auto differenceAverage = std::chrono::microseconds(0);
            for (auto &difference : differences)
            {
                differenceAverage += difference.difference;
            }
            differenceAverage /= CACHED_DIFFERENCE_COUNT;

            setSyncDifference(std::chrono::duration_cast<std::chrono::milliseconds>(differenceAverage));

            if (!clocksSynced)
            {
                clocksSynced = true;
                std::cout << "clocks synced with server" << std::endl;
                sendProtocolMessage(ProtocolMessageType::CLOCKS_SYNCED);
                clocksSyncedCV.notify_all();
            }
            std::this_thread::sleep_for(PING_INTERVAL);
        }
    }
}

void sendInitInfo()
{
    std::lock_guard<std::mutex> lock(sendPacketMutex);

    tcpClient->send(PROTOCOL_PREAMBLE, (sizeof PROTOCOL_PREAMBLE) - 1);

    std::ifstream uuidReadStream("id.txt");

    std::string uuid = "";

    if (uuidReadStream)
    {
        std::stringstream buffer;
        buffer << uuidReadStream.rdbuf();
        uuidReadStream.close();
        uuid = buffer.str();
    }

    uint8_t header[1];
    header[0] = uuid.length();
    if (uuid.length() > 50)
    {
        std::cout << "uuid file contents too long" << std::endl;
        header[0] = 0;
    }
    tcpClient->send((char *)header, 1);
    if (header[0] != 0)
    {
        tcpClient->send(uuid.c_str(), uuid.length());
    }
}

bool disconnectFromServer()
{
    std::cout << "terminating existing connection" << std::endl;

    {
        std::lock_guard<std::mutex> lock(protocolInitializedMutex);
        protocolInitialized = false;
    }
    {
        std::lock_guard<std::mutex> lock(syncThreadShouldContinueMutex);
        syncThreadShouldContinue = true;
    }
    syncThreadNotificationVariable.notify_one();
    tcpClient->disconnect();
    return true;
}

std::string previousConnectError = "";

// TODO: Implement system where reconnection isn't attempted until everything dependent on the connection has been reset to waiting until connected state
void connectToServer()
{
    std::cout << "trying to connect to server" << std::endl;
    while (true)
    {
        try
        {
            tcpClient->connect(serverAdderss, serverPort);
            break;
        }
        catch (TCPClient::ConnectException &e)
        {
            std::string connectError = e.what();
            if (connectError != previousConnectError)
            {
                std::cerr << "failed to connect to server: " << connectError << std::endl;
                previousConnectError = connectError;
            }
        };
        std::this_thread::sleep_for(std::chrono::seconds(1));

        // TODO: implement way of shutting down
    }

    std::cout << "connected to server. Sending init info" << std::endl;
    sendInitInfo();
    std::cout << "sent init info" << std::endl;

    latestRecievedMessageTime = std::chrono::steady_clock::now();
}

void connectionThreadFunction()
{
    char tcpDataBuffer[MAX_MESSAGE_LENGTH];

    uint16_t dataBufferLength = 0;
    bool needsConnecting = true;

    while (true)
    {
        if (needsConnecting)
        {
            needsConnecting = false;
            connectToServer();
        }

        int bytesRead;
        try
        {
            bytesRead = tcpClient->receive(tcpDataBuffer + dataBufferLength, MAX_MESSAGE_LENGTH - dataBufferLength);
        }
        catch (TCPClient::NotConnectedException &e)
        {
            needsConnecting = true;
            continue;
        }
        catch (TCPClient::ReceiveException &e)
        {
            std::cerr << "failed to read message from socket: " << e.what() << std::endl;

            needsConnecting = true;
            continue;
        }

        if (bytesRead == 0) // disconnect
        {
            // TODO: handle random disconnects by distinguishing between them and the ones that are caused by actual wantend disconnects
            disconnectFromServer();

            std::cout << "disconnected suddenly. Waiting 1 second before reconnect attempt" << std::endl;
            std::this_thread::sleep_for(std::chrono::seconds(1));
            continue;
        }
        // else if (bytesRead < 0) // general error
        // {
        //     std::cerr << "failed to read message from socket" << std::endl;
        //     continue;
        // }

        latestRecievedMessageTime = std::chrono::steady_clock::now();

        dataBufferLength += bytesRead;

        protocolInitializedMutex.lock();
        if (!protocolInitialized)
        {
            protocolInitializedMutex.unlock();
            // init data format: [id length: 1 byte][id: id length bytes]
            // if id length is 0, then no id is sent and current is valid

            if (dataBufferLength < 1)
                continue;

            uint8_t uuidLength = tcpDataBuffer[0];

            if (dataBufferLength < uuidLength + 1)
                continue;

            std::string uuid(reinterpret_cast<char *>(tcpDataBuffer + 1), uuidLength);

            if (uuidLength > 0)
            {
                std::ofstream uuidWriteStream("id.txt");
                uuidWriteStream << uuid;
                uuidWriteStream.close();
                std::cout << "got new uuid: " << uuid << std::endl;
            }
            else
            {
                std::cout << "current uuid valid" << std::endl;
            }

            {
                // nothing should be able to send packets before this point, but leaving the lock for the sake of consistency
                std::lock_guard<std::mutex> lock(sendPacketMutex);
                uint8_t header[1];
                header[0] = 0;
                try
                {
                    tcpClient->send((char *)header, 1);
                }
                catch (TCPClient::NotConnectedException &e)
                {
                    std::cerr << "failed to send init data to server: " << e.what() << std::endl;
                    disconnectFromServer();
                    continue;
                }
            }

            dataBufferLength -= uuidLength + 1;

            {
                std::lock_guard<std::mutex> lock(protocolInitializedMutex);
                protocolInitialized = true;
            }
            syncThreadNotificationVariable.notify_one();
        }
        protocolInitializedMutex.unlock();

        while (true) // loop for every packet that can be read from the revieve buffer
        {
            if (dataBufferLength < HEADER_SIZE) // header not yet read. thus can't yet read the message length
            {
                break;
            }

            uint16_t currentPacketLength = ntohs(reinterpret_cast<uint16_t *>(tcpDataBuffer + 1)[0]);
            if (currentPacketLength < HEADER_SIZE)
            {
                std::cerr << "Invalid packet size. Packet size claimed smaller than header size: " << currentPacketLength << std::endl;
                std::cerr << "Disconnecting because of invalid packets" << std::endl;

                disconnectFromServer();
                break;
            }

            if (dataBufferLength < currentPacketLength) // packet not fully received
            {
                break;
            }

            switch (tcpDataBuffer[0])
            {
                case ProtocolMessageType::PING:
                {
                    try
                    {
                        sendProtocolMessage(ProtocolMessageType::PONG, nullptr, 0);
                    }
                    catch (TCPClient::NotConnectedException &e)
                    {
                    }

                    break;
                }
                case ProtocolMessageType::PONG:
                {
                    if (currentPacketLength != HEADER_SIZE + 8)
                    {
                        std::cerr << "Pong message length is invalid: " << currentPacketLength << std::endl;
                        break;
                    }
                    latestRecievedServerTime = std::chrono::milliseconds(ntohll(reinterpret_cast<uint64_t *>(tcpDataBuffer + HEADER_SIZE)[0]));

                    clockSyncMessageHighResRecieveTime = std::chrono::steady_clock::now();

                    {
                        std::lock_guard<std::mutex> lock(syncThreadShouldContinueMutex);
                        syncThreadShouldContinue = true;
                    }
                    syncThreadNotificationVariable.notify_one();
                    break;
                }
                case ProtocolMessageType::MESSAGE:
                {
                    RecieveMessageType messageType = static_cast<RecieveMessageType>((tcpDataBuffer + HEADER_SIZE)[0]);

                    onMessageReceived(messageType, reinterpret_cast<unsigned char *>(tcpDataBuffer) + HEADER_SIZE + 1, currentPacketLength - HEADER_SIZE - 1);
                    break;
                }
                default:
                {
                    std::cerr << "Invalid message type: " << tcpDataBuffer[0] << std::endl;
                    break;
                }
            }

            dataBufferLength -= currentPacketLength;
            memmove(tcpDataBuffer, tcpDataBuffer + currentPacketLength, dataBufferLength);
        }
    }
}

} // namespace

void networking::start(std::string host, std::string port)
{
    serverAdderss = host;
    serverPort = port;
    tcpClient = new TCPClient();

    connectionThread = std::thread(connectionThreadFunction);
    clockSyncThread = std::thread(clockSyncThreadFunction);
}

void networking::sendMessage(SendMessageType type, const unsigned char *message, uint16_t length)
{
    {
        std::lock_guard<std::mutex> lk(protocolInitializedMutex);
        if (!protocolInitialized)
            throw NotConnectedException("not connected to server");
    }
    std::lock_guard<std::mutex> lock(sendPacketMutex);

    int totalMessageLength = HEADER_SIZE + 1 + length;
    uint8_t header[HEADER_SIZE + 1];
    header[0] = ProtocolMessageType::MESSAGE;
    uint16_t *header_messageLength = (uint16_t *)&header[1];
    *header_messageLength = htons(totalMessageLength);
    header[3] = static_cast<uint8_t>(type);

    tcpClient->send((char *)&header, HEADER_SIZE + 1);
    if (length != 0)
    {
        tcpClient->send((char *)message, length);
    }
}

void networking::sendMessage(SendMessageType type)
{
    sendMessage(type, nullptr, 0);
}
