#include <unistd.h>

#include <algorithm>
#include <chrono>
#include <cstring>
#include <iostream>
#include <string>
#include <thread>

#include "../timing.hpp"
#include "playbackSystem.hpp"

#define MUTEX_WAIT_BUFFER std::chrono::milliseconds(50)

static std::mutex sleepingMutex;
static std::condition_variable sleepingConditionVariable;

static bool playbackSystemShouldShutdown = false;

static std::mutex threadSafetyMutex;

static bool settingTimedFunction = false;
static bool playbackSystemThreadNext = false;

struct EventInfo
{
    std::function<void()> callback;
    std::chrono::milliseconds eventTime;

    bool operator<(EventInfo const &other)
    {
        return eventTime < other.eventTime;
    }
    bool operator<(std::chrono::milliseconds const &other)
    {
        return eventTime < other;
    }
};

static std::vector<EventInfo> eventList;

void addEvent(std::function<void()> callback, std::chrono::milliseconds eventTime)
{
    EventInfo eventInfo;
    eventInfo.callback = callback;
    eventInfo.eventTime = eventTime;
    bool addedAnything = false;

    threadSafetyMutex.lock();
    auto position = std::lower_bound(eventList.begin(), eventList.end(), eventTime);
    eventList.insert(position, eventInfo);
    threadSafetyMutex.unlock();

    settingTimedFunction = true;
    sleepingConditionVariable.notify_one();
}

void runTimingSystemLoop()
{
    while (true)
    {
        if (eventList.size() == 0)
        {
            std::unique_lock<std::mutex> lk(sleepingMutex);
            threadSafetyMutex.unlock();
            sleepingConditionVariable.wait(lk, [&] { return (playbackSystemShouldShutdown || settingTimedFunction); });
            threadSafetyMutex.lock();

            if (playbackSystemShouldShutdown)
            {
                threadSafetyMutex.unlock();
                break;
            }
        }

        settingTimedFunction = false;

        auto nextEventStartTime = std::chrono::steady_clock::time_point(std::chrono::microseconds(serverTimeToLocalTime(eventList[0].eventTime)));

        std::unique_lock<std::mutex> lk(sleepingMutex);

        if (nextEventStartTime - MUTEX_WAIT_BUFFER > std::chrono::steady_clock::now())
        {
            playbackSystemThreadNext = false;
            threadSafetyMutex.unlock();
            bool sleepWasCancelled = sleepingConditionVariable.wait_until(
                lk, nextEventStartTime - MUTEX_WAIT_BUFFER, [&] { return (playbackSystemShouldShutdown || settingTimedFunction); }
            );

            threadSafetyMutex.lock();

            if (playbackSystemShouldShutdown)
            {
                threadSafetyMutex.unlock();
                break;
            }
            if (sleepWasCancelled)
            {
                continue;
            }
        }
        else
        {
            std::cerr << "timed action might have been called late" << std::endl;
        }

        while (nextEventStartTime > std::chrono::steady_clock::now())
        {
            std::this_thread::sleep_for(std::chrono::microseconds(100));
        }

        eventList[0].callback();

        eventList.erase(eventList.begin());
    }
}

void shutdown()
{
    threadSafetyMutex.lock();
    playbackSystemShouldShutdown = true;
    sleepingConditionVariable.notify_one();
    threadSafetyMutex.unlock();
}
